var request = require('request');

// API.
//
// @param		Object			client			The client instance
// @param		Object			options			The request options
function API (client, options) {
	this.client = client;
	this.options = options || {};
	this.sessionEnabled = options.session !== false;
	this.geolocationEnabled = options.geolocation !== false;
	this.aborted = false;
	this.errorCount = 0;
	this.maxErrors = 1;
	this.makeRequest();
}

// Wrapper for making a request that ensures a session. The task queue ensures
// that multiple simultaneous requests are queued for session retrieval.
//
// @return		Object						This instance for chainability
API.prototype.makeRequest = function () {
	var self = this;
	if (this.sessionEnabled) {
		this.client.taskQueues.obtainSession.invoke(function (err) {
			self[!err ? '_makeRequest' : 'abort']();
		});
	} else {
		this._makeRequest();
	}
	return this;
};

// Private method for making a request after a session has been obtained.
//
// @return		void
API.prototype._makeRequest = function () {
	if (this.aborted || (this.sessionEnabled && !this.client.session.get('token'))) return;
	var options = this.options;
	var callback = options.callback;
	var qs = {};
	var self = this;
	var xhr;

	options.json = true;
	options.qs = options.qs || {};
	options.headers = options.headers || {};

	// Set token and signature request headers.
	if (this.sessionEnabled) {
		var sessionHeaders = this.client.session.getHeaders();
		for (var sessionHeader in sessionHeaders) {
			options.headers[sessionHeader] = sessionHeaders[sessionHeader];
		}
	}

	// Override callback.
	options.callback = function (err, res, body) {
		var status = res.statusCode;
		var isSuccess = !err && status >= 200 && status < 300 || status === 304;
		var successCallback = function () {
			if (callback) callback(err, res, body);
		};
		var errorCallback = function (_err) {
			if (callback) callback(_err || err, res, body);
		};
		var createSession = function () {
			self.client.taskQueues.createSession.invoke(function (err) {
				if (!err) {
					self._makeRequest();
				} else {
					done();
				}
			});
		};
		var tryAgain = function () {
			setTimeout(function () {self._makeRequest(); }, retryAfter * 1000);
		};

		if (isSuccess) {
			// Handle changes in session.
			if (self.sessionEnabled) {
				var currentToken = self.client.session.get('token');
				var returnedToken = xhr && xhr.getHeader('X-Token');
				if (returnedToken && returnedToken !== currentToken) {
					self.client.session.set({ token: returnedToken }, true);
				}
			}
			successCallback();
		} else {
			var code = body && body.code;
			var retryAfter = xhr && xhr.getHeader('Retry-After');
			var error = new Error('ApiError');
			error.name = self.errors[code] || 'UNKNOWN_ERROR';

			// Don't try to fix error if the error count has exceeded the
			// maximum errors allowed.
			if (self.errorCount >= self.maxErrors) {
				return errorCallback(error);
			}

			// Increment the error count.
			self.errorCount++;

			// Session errors.
			if (self.sessionEnabled && (code === 1101 || code === 1108 || code === 1300 || code === 1301)) {
				createSession();

			// Temporary error - try again.
			} else if (code == 2015 && isFinite(retryAfter) && retryAfter <= 1) {
				tryAgain();

			// Other error.
			} else {
				errorCallback(error);
			}
		}
	};

	// Include geolocation if set and enabled.
	if (this.geolocationEnabled) {
		var latitude = this.client.geolocation.get('latitude');
		var longitude = this.client.geolocation.get('longitude');
		var radius = this.client.geolocation.get('radius');
		var sensor = this.client.geolocation.get('sensor');

		if ('r_lat' in options.qs) {
			qs.r_lat = options.qs.r_lat;
		} else if (latitude) {
			qs.r_lat = latitude;
		}
		if ('r_lng' in options.qs) {
			qs.r_lng = options.qs.r_lng;
		} else if (longitude) {
			qs.r_lng = longitude;
		}
		if ('r_radius' in options.qs) {
			qs.r_radius = options.qs.r_radius;
		} else if (radius) {
			qs.r_radius = radius;
		}
		if ('r_sensor' in options.qs) {
			qs.r_sensor = options.qs.r_sensor;
		} else if (sensor) {
			data.r_sensor = sensor;
		}
	}

	// Apply parameters so that API-specific parameters come first in the query string.
	for (var key in options.qs) qs[key] = options.qs[key];
	options.qs = qs;

	// Prefix URL.
	if (options.url.substring(0, 1) === '/') {
		options.url = this.client.options.apiURL + options.url;
	}

	// Make the request.
	xhr = request(options);
};

// Abort request.
//
// @return		void
API.prototype.abort = function () {
	this.aborted = true;
};

// Hash of errors we need to check for in unsuccessful requests.
API.prototype.errors = {
	// Session errors.
	1100: 'SESSION_ERROR',
	1101: 'SESSION_TOKEN_EXPIRED',
	1102: 'SESSION_INVALID_API_KEY',
	1103: 'SESSION_MISSING_SIGNATURE',
	1104: 'SESSION_INVALID_SIGNATURE',
	1105: 'SESSION_TOKEN_NOT_ALLOWED',
	1106: 'SESSION_MISSING_ORIGIN',
	1107: 'SESSION_MISSING_TOKEN',
	1108: 'SESSION_INVALID_TOKEN',

	// Authentication.
	1200: 'AUTHENTICATION_ERROR',
	1201: 'AUTHENTICATION_INVALID_CREDENTIALS',
	1202: 'AUTHENTICATION_NO_USER',
	1203: 'AUTHENTICATION_EMAIL_NOT_VERIFIED',

	// Authorization.
	1300: 'AUTHORIZATION_ERROR',
	1301: 'AUTHORIZATION_ACTION_NOT_ALLOWED',

	// Missing information.
	1400: 'INFO_REQUEST_INVALID',
	1401: 'INFO_MISSING_GEOLOCATION',
	1402: 'INFO_MISSING_RADIUS',
	1411: 'INFO_MISSING_AUTHENTICATION',
	1431: 'INFO_MISSING_EMAIL',
	1432: 'INFO_MISSING_BIRTHDAY',
	1433: 'INFO_MISSING_GENDER',
	1434: 'INFO_MISSING_LOCALE',
	1435: 'INFO_MISSING_NAME',
	1440: 'INFO_RESOURCE_NOT_FOUND',

	// Invalid information.
	1500: 'INFO_INVALID',
	1501: 'INFO_INVALID_RESOURCE_ID',
	1530: 'INFO_RESOURCE_DUPLICATION',
	1566: 'INFO_INVALID_BODY_DATA',

	// Internal corruption of data.
	2000: 'INTERNAL_INTEGRITY_ERROR',
	2010: 'INTERNAL_SEARCH_ERROR',
	2015: 'INTERNAL_NON_CRITICAL_ERROR',

	// Misc.
	4000: 'MISC_ACTION_NOT_EXISTS'
};

module.exports = API;
