var util = require('util');
var events = require('events');
var Session = require('./session');
var Geolocation = require('./geolocation');
var API = require('./api');
var TaskQueue = require('./task_queue');

// Main class to interface with.
//
// The configuration:
// - apiKey				The API key for requests.
// - apiSecret			The API secret for requests.
// - apiURL				The main URL for API requests (optional).
// - session			The session to use instead of looking in storage (optional).
// - geolocation		The default geolocation (optional).
function Client (options) {
	this.options = util._extend(util._extend({}, this.defaults), options);

	// You can retrieve an API key and secret at:
	// http://etilbudsavis.dk/developers/
	var error = new Error('ClientInitializeError');
	if (this.options.apiKey == null) {
		error.name = 'API_KEY_MISSING';
		throw error;
	}
	if (this.options.apiSecret == null) {
		error.name = 'API_SECRET_MISSING';
		throw error;
	}

	this.hasRequestSession = !!(this.options.request && this.options.request.session);
	this.setupSession(this.getPersistentData('etasession') || this.options.session);
	this.setupGeolocation(this.getPersistentData('etageolocation') || this.options.geolocation);
	this.setupTaskQueues();
}

Client.prototype = new events.EventEmitter();

// Get persistent data from session.
//
// @param		String			key				The data key
// @param		Mixed			defaultData		The data to return if no data was found
// @return		Mixed							The found data
Client.prototype.getPersistentData = function (key, defaultData) {
	if (this.hasRequestSession) {
		return this.options.request.session[key] || defaultData;
	}
	return defaultData;
};

// Set persistent data to session.
//
// @param		String			key				The data key
// @param		Mixed			data			The data
// @return		void
Client.prototype.setPersistentData = function (key, data) {
	if (this.hasRequestSession) {
		this.options.request.session[key] = data;
	}
};

// Clear persistent data from session.
//
// @param		String			key				The data key
// @return		void
Client.prototype.clearPersistentData = function (key) {
	if (this.hasRequestSession) {
		delete this.options.request.session[key];
	}
};

// Setup session.
//
// @param		Object			session			The predefined session to trump storage
// @return		void
Client.prototype.setupSession = function (session) {
	var self = this;

	this.session = new Session(session, this);
	this.session.on('_change', function (attributes) {
		self.emit('session:change', attributes);
		self.setPersistentData('etasession', attributes);
	});
	this.session.on('_reset', function () {
		self.clearPersistentData('etasession');
	});
};

// Setup geolocation.
//
// @param		Object			geolocation		The predefined geolocation to trump storage
// @return		void
Client.prototype.setupGeolocation = function (geolocation) {
	var self = this;

	this.geolocation = new Geolocation(geolocation);
	this.geolocation.on('_change', function (attributes) {
		self.emit('geolocation:change', attributes);
		self.setPersistentData('etageolocation', attributes);
	});
	this.geolocation.on('_reset', function () {
		self.clearPersistentData('etageolocation');
	});
};

// Setup task queues.
//
// @return		void
Client.prototype.setupTaskQueues = function () {
	var self = this;

	this.taskQueues = {
		obtainSession: new TaskQueue(function (callback) {
			self.session.obtain(callback);
		}),
		createSession: new TaskQueue(function (callback) {
			self.session.reset();
			self.session.create(callback);
		})
	};
};

// Make API requests.
//
// @param		Object			options			The options
// @return		Object							The API instance
Client.prototype.api = function (options) {
	return new API(this, options);
};

// Default options.
Client.prototype.defaults = {
	apiURL: 'https://api.etilbudsavis.dk'
};

Client.middleware = function (options) {
	options = options || {};
	return function (req, res, next) {
		options.request = req;
		options.response = res;
		req.eta = new Client(options);
		next();
	};
};

Client.requireAuth = function (options, actions) {
	return function (req, res, next) {
		if (!req.eta) {
			ETA.middleware(options)(req, res, afterNew);
		} else {
			afterNew();
		}

		function afterNew () {
			req.eta.taskQueues.obtainSession.invoke(function (err) {
				var error = null;
				if (err || !req.eta.session.isUserAllowed(actions)) {
					error = new Error('UnauthorizedError');
					error.name = 'USER_UNAUTHORIZED';
				}
				next(error);
			});
		}
	};
};

module.exports = Client;
