var util = require('util');
var events = require('events');
var crypto = require('crypto');
var Permissions = require('./permissions');

// Session management.
//
// @param		Object			attributes		The attributes
// @param		Object			client			The client instance
function Session (attributes, client) {
	this.attributes = attributes || {};
	this.client = client;
	this.url = this.client.options.apiURL + '/v2/sessions';
}

Session.prototype = new events.EventEmitter();

// Get attribute key value.
//
// @param		String			key				The key
// @return		Mixed							The attribute key value
Session.prototype.get = function (key) {
	return this.attributes[key];
};

// Set attributes.
//
// @param		Object			attributes		The session attributes
// @param		Boolean			silent			True if no events should be propagated
// @return		void
Session.prototype.set = function (attributes, silent) {
	var changes = false;
	for (var key in attributes) {
		if (this.attributes[key] !== attributes[key]) {
			this.attributes[key] = attributes[key];
			changes = true;
		}
	}
	if (changes) {
		var clone = this.clone();
		this.emit('_change', clone);
		if (!silent) this.emit('change', clone);
	}
};

// Get request headers needed for API requests.
//
// @return		Object							The headers
Session.prototype.getHeaders = function () {
	var headers = {};
	var token = this.get('token');
	if (token) {
		headers['X-Token'] = token;
		if (this.client.options.apiSecret) headers['X-Signature'] = this.getSignature();
	}
	return headers;
};

// Get signature.
//
// @return		String							The signature
Session.prototype.getSignature = function () {
	var sha256 = crypto.createHash('sha256');
	sha256.update([this.client.options.apiSecret, this.get('token')].join(''), 'utf8');
	return sha256.digest('hex');
};

// Has session expired?
//
// @return		Boolean							True if session has expired
Session.prototype.hasExpired = function () {
	if (this.get('token') && this.get('expires')) {
		return new Date().getTime() >= Date.parse(this.get('expires'));
	} else {
		return true;
	}
};

// Has an user been attached to the session?
//
// @return		Boolean							True if an user has been attached
Session.prototype.isUserAttached = function () {
	return !!this.get('user');
};

// Get the user id attached to the session.
//
// @return		Number							The user id
Session.prototype.getUserId = function () {
	if (this.isUserAttached()) return this.get('user').id;
};

// Clone session attributes.
//
// @return		Object							The session attributes
Session.prototype.clone = function () {
	return util._extend({}, this.attributes);
};

// Reset attributes.
//
// @param		Boolean			silent			True if no events should be propagated
// @return		void
Session.prototype.reset = function (silent) {
	this.attributes = {};
	this.emit('_reset');
	if (!silent) this.emit('reset');
};

// Check if user is allowed.
//
// @param		Array			actions			Required permission actions
// @return		Boolean							True if user is allowed
Session.prototype.isUserAllowed = function (actions) {
	var user = this.get('user');
	if (!user) return false;
	if (!actions || actions.length === 0) return true;
	return this.allowsAll(actions);
};

// Check if permissions stored in session allows given action.
//
// @param		String			action			The action
// @return		Boolean							True if action is allowed
Session.prototype.allows = function (action) {
	// Don't allow if no permissions.
	var permissions = this.get('permissions');
	if (!permissions) return false;

	for (var name in permissions) {
		var permissionCollection = new Permissions(permissions[name], name);
		if (permissionCollection.allows(action)) {
			return true;
		}
	}
	return false;
};

// Check if permissions stored in session allows all given actions.
//
// @param		Array			actions			The actions
// @return		Boolean							True if actions are allowed
Session.prototype.allowsAll = function (actions) {
	for (var i = 0, len = actions.length; i < len; i++) {
		if (!this.allows(actions[i])) {
			return false;
		}
	}
	return true;
};

// Obtain session attributes.
//
// @param		Function		callback		The callback
// @return		void
Session.prototype.obtain = function (callback) {
	var token = this.get('token');
	var expires = this.get('expires');
	var now = new Date().getTime();
	var day = 1000 * 60 * 60 * 24 * 2;
	var self = this;

	function done (err, isNew) {
		if (!err) {
			callback(null, self.clone(), isNew);
		} else {
			callback(err);
		}
	}

	if (token && expires) {
		if (now + day <= Date.parse(expires)) {
			done(null, false);
		} else {
			this.renew(function (err) {
				if (!err) {
					done(err, true);
				} else {
					self.create(function (err) {
						done(err, true);
					});
				}
			});
		}
	} else {
		this.create(function (err) {
			done(err, true);
		});
	}
};

// Create session via the API.
//
// @param		Function		callback		The callback
// @return		void
Session.prototype.create = function (callback) {
	var self = this;
	this.client.api({
		method: 'post',
		url: this.url,
		qs: {
			api_key: this.client.options.apiKey
		},
		session: false,
		geolocation: false,
		callback: function (err, res, body) {
			if (!err) {
				self.set(body);
				callback(null, self.clone());
			} else {
				callback(err);
			}
		}
	});
};

// Update session via the API.
//
// @param		Function		callback		The callback
// @return		void
Session.prototype.update = function (callback) {
	var self = this;
	var error;

	if (!this.get('token')) {
		error = new TypeError('SessionError');
		error.name = 'TOKEN_MISSING';
		return callback(error);
	}

	this.client.api({
		url: this.url,
		headers: this.getHeaders(),
		session: false,
		geolocation: false,
		callback: function (err, res, body) {
			if (!err) {
				self.set(body);
				callback(null, self.clone());
			} else {
				callback(err);
			}
		}
	});
};

// Renew session via the API.
//
// @param		Function		callback		The callback
// @return		void
Session.prototype.renew = function (callback) {
	var self = this;
	var error;

	if (!this.get('token')) {
		error = new TypeError('SessionError');
		error.name = 'TOKEN_MISSING';
		return callback(error);
	}

	this.client.api({
		method: 'put',
		url: this.url,
		headers: this.getHeaders(),
		session: false,
		geolocation: false,
		callback: function (err, res, body) {
			if (!err) {
				self.set(body);
				callback(null, self.clone());
			} else {
				callback(err);
			}
		}
	});
};

// Attach user to session via the API.
//
// @param		Object			credentials		The user credentials
// @param		Function		callback		The callback
// @return		void
Session.prototype.attachUser = function (credentials, callback) {
	var self = this;
	var error;

	if (!this.get('token')) {
		error = new TypeError('SessionError');
		error.name = 'TOKEN_MISSING';
		return callback(error);
	}

	this.detachUser(function () {
		self.client.api({
			method: 'put',
			url: self.url,
			headers: self.getHeaders(),
			qs: credentials,
			session: false,
			geolocation: false,
			callback: function (err, res, body) {
				if (!err) {
					self.set(body);
					callback(null, self.clone());
				} else {
					callback(err);
				}
			}
		});
	});
};

// Detach user from session via the API.
//
// @param		Function		callback		The callback
// @return		void
Session.prototype.detachUser = function (callback) {
	var self = this;
	var error;

	if (!this.get('token')) {
		error = new TypeError('SessionError');
		error.name = 'TOKEN_MISSING';
		return callback(error);
	}

	this.client.api({
		method: 'put',
		url: this.url,
		headers: this.getHeaders(),
		qs: {
			email: ''
		},
		session: false,
		geolocation: false,
		callback: function (err, res, body) {
			if (!err) {
				self.reset(true);
				self.set(body);
				callback(null, self.clone());
			} else {
				callback(err);
			}
		}
	});
};

module.exports = Session;
