// Queue a task.
//
// @param		Function		task			The task to be worked on
//
// Example:
// var taskQueue = new TaskQueue(function (resolve, reject) {
//     setTimeout(function () {
//         resolve();
//     }, 1000);
// });
//
// taskQueue.invoke()
// .done(function () { console.log('done'); })
// .fail(function () { console.log('fail'); });
//
// taskQueue.invoke()
// .done(function () { console.log('done second'); })
// .fail(function () { console.log('fail second'); });
//
// ...
//
// n
function TaskQueue (task) {
	// A task has to be a function.
	this.task = typeof task === 'function' ? task : function () {};
	this.running = false;
	this.queue = [];
}

// Resolve all callbacks in the queue stack and run again.
//
// @return		void
TaskQueue.prototype.resolve = function () {
	for (var key in this.queue) this.queue[key](null);
	this.stopRun();
	this.run();
};

// Reject all callbacks in the queue stack and run again.
//
// @return		void
TaskQueue.prototype.reject = function (err) {
	for (var key in this.queue) this.queue[key](err);
	this.stopRun();
	this.run();
};

// Go through all callbacks in the queue stack if not already running.
// It will keep calling itself until the queue is empty.
//
// @return		void
TaskQueue.prototype.run = function () {
	if (this.running || this.queue.length === 0) return;
	var self = this;
	this.running = true;
	this.task(function (err) {
		if (!err) {
			self.resolve();
		} else {
			self.reject(err);
		}
	});
};

// Reset the queue stack and stop running.
//
// @return		void
TaskQueue.prototype.stopRun = function () {
	this.queue = [];
	this.running = false;
};

// Use this to async subscribe to the task.
//
// @param		Function		callback		The callback for when the task has resolved
// @return		void
TaskQueue.prototype.invoke = function (callback) {
	this.queue.push(callback);
	this.run();
};

module.exports = TaskQueue;
